#include "epplet.h"
#include <string.h>
#include <dirent.h>

/* Should you have problems with missing battery symbols please enable
 * the next line and set your path accordingly.
 */

//#define EROOT "/usr/share/enlightenment"

//#define DEBUG_PRINT
#define DEBUG_FILE "/tmp/acpi.epplet.debug"

#define MAX_BATS 4

static char * state_file[MAX_BATS];
static char * info_file[MAX_BATS];
static char * ac_file;

int		num_bats = 0;
int		bat_val[MAX_BATS];
int		bat_perc[MAX_BATS];
int		bat_rate[MAX_BATS];
int		bat_max[MAX_BATS];
int		ac_stat = 0;
int		time_val = 0;
int		time_counts = 0;
int		required_diff = 10;
int		virt_bat_oldval = 0;
Epplet_gadget	b_close, b_help, image, label;

const int timer_intervall = 15;

static void cb_timer		(void *data);
static void cb_close		(void *data);
static void cb_in		(void *data, Window w);
static void cb_out		(void *data, Window w);
static void cb_help		(void *data);
static void ProbePowerInterface	(void);

static void
cb_timer (void *data)
{
	static FILE *f = NULL;
	char s[127];
	char filesfound = 0;
	int rate_display[MAX_BATS];
	int i;

	for (i=0; i<MAX_BATS; ++i) {
		rate_display[i] = 0;
		bat_val[i] = 0;
		bat_perc[i] = 0;
		bat_rate[i] = 0;
		bat_max[i] = 0;
	}

	/* Discover any new batteries */
	ProbePowerInterface();
	
	if ( !(num_bats && ac_file) ) {
		// Check if at least /proc/acpi is present
		f = fopen( "/proc/acpi", "r");
		if (f) {
			if (num_bats)
				Epplet_change_label(label, "no info av.");
			else
				Epplet_change_label(label, "no batt av.");
			fclose(f);
#ifdef DEBUG_PRINT
		} else {
			f = fopen (DEBUG_FILE, "a");
			if (f) {
				fprintf(f,"error: can not open /proc/acpi.\n");
				fclose (f);
			}
#endif
		}
		
		Epplet_timer (cb_timer, NULL, timer_intervall, "TIMER");
		return;
	}

	f = fopen(ac_file, "r");
	if (f) {
		// Check whether the power supply is connected
		char s1[32], s2[32];
		fgets(s, 127, f);
		sscanf(s, "%s %s", s1, s2);
		if (!strcasecmp(s2, "off-line"))
			ac_stat = 0;
		else {
			if (!strcasecmp(s2, "on-line"))
				ac_stat = 1;
			else
				bat_val[i] = 0;
		}
		fclose(f);
		filesfound++;
#ifdef DEBUG_PRINT
	} else {
		f = fopen (DEBUG_FILE, "a");
		if (f) {
			fprintf(f,"error: can not open %s\n",ac_file);
			fclose (f);
		}
#endif
	}

	for (i=0; i<num_bats; ++i) {
		f = fopen (state_file[i], "r");
		if (f) {
			// Get the current loading current and the capacity left
			char s1[32], s2[32], s3[32], s4[32];
			int need_volts = 0;
			int volts = 0;

			fgets(s, 127, f);
			while (!feof(f)) {
				sscanf(s, "%s %s %s %s", s1, s2, s3, s4);
				if (!strcasecmp (s1, "Remaining")) {
					bat_val[i] = atoi (s3);
				}
				if (!strcasecmp (s2, "Rate:")) {
					rate_display[i] = bat_rate[i] = atoi(s3);
					s4[2] = 0;
					// if it is in watts, then we will calculate amperes...
					if (!strcasecmp (s4, "mw")) {
						need_volts = 1;
					}
				}
				if (!strcasecmp(s2, "voltage:")) {
					volts = atoi (s3);
				}
				fgets(s, 127, f);
 			}
			fclose (f);
			
			if (need_volts && volts) {
				// convert watts to amperes
				rate_display[i] *= 1000;
				rate_display[i] /= volts;
			}
			
			filesfound++;
#ifdef DEBUG_PRINT
		} else {
			f = fopen (DEBUG_FILE, "a");
			if (f) {
				fprintf(f,"error: can not open %s\n",state_file[i]);
				fclose (f);
			}
#endif
		}

		f = fopen (info_file[i], "r");
		if (f) {
			// Get the maximum capacity of the battery
			char s1[32], s2[32], s3[32], s4[32];

			fgets (s, 127, f);
			while (!feof (f)) {
				sscanf(s, "%s %s %s %s", s1, s2, s3, s4);
				if (!strcasecmp (s2, "Full")) {
					bat_max[i] = atoi (s4);
				}
				fgets (s, 127, f);
			}
			fclose (f);
			
			filesfound++;
#ifdef DEBUG_PRINT
		} else {
			f = fopen (DEBUG_FILE, "a");
			if (f) {
				fprintf(f,"error: can not open %s\n",info_file[i]);
				fclose (f);
			}
#endif
		}
	}

	if (filesfound >= 2*num_bats+1) {  // 1 ac file and 2 info files per battery
		for (i=0; i<num_bats; ++i) {
			if ( (bat_val[i] == 0) || (bat_max[i] == 0) )
				bat_perc[i] = 100;
			else
				bat_perc[i] = bat_val[i] * 100 / bat_max[i];
		}

	    // now create all the strings for the labels..
	    {
		int hours=0, minutes=0, percent=0, rate=0, j;
		int virt_bat_max = 0, virt_bat_rate = 0, virt_bat_val = 0;

#ifdef DEBUG_PRINT
		FILE *f;
		f = fopen (DEBUG_FILE, "a");
		if (f) {
			fprintf(f,"num bats found: %d\n",num_bats);
		}
#endif
		for (j=0; j<num_bats; ++j) {
		//create a "virtual battery" that is the sum of all batteries in system
			virt_bat_max  += bat_max[j];
			virt_bat_rate += bat_rate[j];
			virt_bat_val  += bat_val[j];

			rate += rate_display[j];

#ifdef DEBUG_PRINT
			if (f) {
				fprintf(f,"bat_perc[%d]=%d\n",j,bat_perc[j]);
				fprintf(f,"bat_val[%d]=%d\n",j,bat_val[j]);
				fprintf(f,"bat_max[%d]=%d\n",j,bat_max[j]);
				fprintf(f,"bat_rate[%d]=%d\n",j,rate_display[j]);
			}
#endif
		}

		if (ac_stat == 1) {
			// Ok, try to calculate the time till full charge
			time_counts++;
			if (virt_bat_oldval == 0) {
				time_counts = 0;
				time_val = 0;
				virt_bat_oldval = virt_bat_val;
				required_diff = 10;
			}
			if (virt_bat_oldval != 0) {
				if (virt_bat_val - virt_bat_oldval > required_diff) {
					time_val = (virt_bat_max - virt_bat_val) 
						/ (virt_bat_val - virt_bat_oldval) 
						* time_counts * timer_intervall;
					virt_bat_oldval = virt_bat_val;
					time_counts = 0;
					required_diff = 20;
				}
			}
		} else if ( (virt_bat_rate != 0) && (virt_bat_val != 0) ) {
			time_val = 3600 * virt_bat_val / virt_bat_rate;
			virt_bat_oldval = 0; // reset this value to get the next on-line phase right
		}

		hours = time_val / 3600;
		minutes = (time_val / 60) % 60;

		if (virt_bat_max)
			percent = virt_bat_val * 100 / virt_bat_max;
		else
			percent = -1;

#ifdef DEBUG_PRINT
		if (f) {
			fprintf(f,"total percent=%d\n",percent);
			fprintf(f,"hours: %d\nminutes: %d\n",hours,minutes);
			fprintf(f,"virt_bat_val: %d\n",virt_bat_val);
			fprintf(f,"virt_bat_max: %d\n",virt_bat_max);
		}
#endif

		if (percent > 0) {
			sprintf (s, "%i%% \n", percent);
			
			if (ac_stat == 1) {
				if ((time_val > 0) && (virt_bat_val < virt_bat_max)) 
					sprintf (s, "%sETA: %i:%02i", s,
						hours, minutes);
				else
					sprintf (s, "%sAC on-line", s);
			} else {
				sprintf (s, "%s%i:%02i", s,
					hours, minutes);
				sprintf (s, "%s  %i.%01iA", s,
					rate / 1000, (rate % 1000) / 100 );
			}

			Epplet_change_label (label, s);

			snprintf(s, sizeof (s), EROOT"/epplet_data/E-AcpiPower/E-AcpiPower-Bat-%i.png", 
				((percent + 5) / 10) * 10);
			//Esnprintf(s, sizeof(s), EROOT"/epplet_data/E-AcpiPower/E-AcpiPower-Bat-%i.png", 
			//	((bat_perc + 5) / 10) * 10);
		} else {
			Epplet_change_label(label, "no info av.");
			snprintf(s, sizeof (s), EROOT"/epplet_data/E-AcpiPower/E-AcpiPower-Bat-%i.png", 
				100);
		}
		Epplet_change_image(image, 44, 24, s);
	    }

#ifdef DEBUG_PRINT
	if (f) {
		fclose(f);
	}
#endif
	}

	Epplet_timer (cb_timer, NULL, timer_intervall, "TIMER");

	data = NULL;
}

static int
test_file (const char *name)
{
	char s1[32], s2[32]; FILE *f;
	f = fopen (name, "r");
	if (!f) return 0;
	fscanf (f, "%s %s", s1, s2);
	if ( (!strcasecmp(s1, "present:")) && (!strcasecmp(s2, "no")) ) {
		fclose (f);
		return 0;
	}
	fclose (f);
	return 1;
}


// Probe the /proc filesystem to discover the
// correct paths for the battery status and the ac status
static void 
ProbePowerInterface (void)
{
	struct dirent *dptr;
	DIR *dirp;
	FILE *f;
	char filename[64];
	int i;

	for (i=0; i<MAX_BATS; ++i) {
		info_file[i] = 0;
		state_file[i] = 0;
	}
	num_bats = 0;

	/* Does /proc/acpi/battery exist ? */
	if ( (dirp = opendir ("/proc/acpi/battery")) == NULL)
		return;

	/* Search for batteries */
	while ((dptr = readdir (dirp))) {
		if (!strncmp (dptr->d_name, ".", 1))
			/* Omit any files starting with a '.' */
			continue;
			
		snprintf (filename, sizeof (filename),
				"/proc/acpi/battery/%s/info", dptr->d_name);

		if (test_file (filename)) {
			/* an info file exists */
			info_file[num_bats] = strdup (filename);
			
			snprintf (filename, sizeof (filename),
					"/proc/acpi/battery/%s/state", dptr->d_name);
			f = fopen (filename, "r");
			if (f) {
				/* a state file exists */
				state_file[num_bats] = strdup (filename);
				fclose (f);
#ifdef DEBUG_PRINT
				f = fopen (DEBUG_FILE, "a");
				if (f) {
					fprintf(f,"using state file: %s\n",state_file[num_bats]);
					fclose (f);
				}
#endif
				num_bats++;
			} else {
				snprintf (filename, sizeof (filename),
						"/proc/acpi/battery/%s/status", dptr->d_name);
				f = fopen (filename, "r");
				if (f) {
					/* a status file exists */
					state_file[num_bats] = strdup (filename);
					fclose (f);
#ifdef DEBUG_PRINT
					f = fopen (DEBUG_FILE, "a");
					if (f) {
						fprintf(f,"using state file: %s\n",state_file[num_bats]);
						fclose (f);
					}
#endif
					num_bats++;
				} else {
					/* no state or status file found */
					free (state_file[num_bats]);
					state_file[num_bats] = NULL;
				}
			}
			
			if (num_bats >= MAX_BATS)
				/* enough is enough */
				break;
		}
	}
	closedir (dirp);

#ifdef DEBUG_PRINT
	f = fopen (DEBUG_FILE, "a");
	if (f) {
		fprintf(f,"found a total of %d batteries\n",num_bats);
		fclose (f);
	}
#endif
	if (!ac_file) {
		/* Does /proc/acpi/ac_adapter exist ? */
		if ( (dirp = opendir ("/proc/acpi/ac_adapter")) == NULL)
			return;

		/* Search for AC adapters */
		while ((dptr = readdir (dirp))) {
			if (!strncmp (dptr->d_name, ".", 1))
				/* Omit any files starting with a '.' */
				continue;

			snprintf (filename, sizeof (filename),
					"/proc/acpi/ac_adapter/%s/state", dptr->d_name);
			
			f = fopen (filename, "r");
			if (f) {
				fclose (f);
				ac_file = strdup (filename);
#ifdef DEBUG_PRINT
				f = fopen (DEBUG_FILE, "a");
				if (f) {
					fprintf(f,"using ac file: %s\n",ac_file);
					fclose (f);
				}
#endif
				break;
			}
		}
		closedir (dirp);
	}
}


static void
cb_close (void *data)
{
	Epplet_unremember();
	Esync();
	data = NULL;
	exit(0);
}

static void
cb_in (void *data, Window w)
{
	if (w == Epplet_get_main_window ()) {
		Epplet_gadget_show(b_close);
		Epplet_gadget_show(b_help);
	}
	return;
	data = NULL;
}

static void
cb_out (void *data, Window w)
{
	if (w == Epplet_get_main_window ()) {
		Epplet_gadget_hide(b_close);
		Epplet_gadget_hide(b_help);
	}
	return;
	data = NULL;
}

static void
cb_help (void *data)
{
	Epplet_show_about("E-AcpiPower");
	return;
	data = NULL;
}

int
main (int argc, char **argv)
{
	atexit (Epplet_cleanup);

	ac_file = NULL;

	Epplet_Init ("E-AcpiPower", "0.9.0", "Enlightenment Laptop ACPI Epplet",
			3, 3, argc, argv, 0);
	b_close = Epplet_create_button (NULL, NULL,
			2, 2, 0, 0, "CLOSE", 0, NULL,
			cb_close, NULL);
	b_help = Epplet_create_button (NULL, NULL,
			34, 2, 0, 0, "HELP", 0, NULL,
			cb_help, NULL);
	Epplet_gadget_show (image = 
			Epplet_create_image
			(2, 2, 44, 24,
			 EROOT"/epplet_data/E-AcpiPower/E-AcpiPower-Bat-100.png"));
	Epplet_gadget_show (label = 
			Epplet_create_label
			(2, 28, "ACPI not\nin Kernel", 1));
	Epplet_register_focus_in_handler (cb_in, NULL);
	Epplet_register_focus_out_handler (cb_out, NULL);
	cb_timer (NULL);
	Epplet_show ();
	Epplet_Loop ();
	return 0;
}
